/**!
   * mongodash v1.1.1
   * https://github.com/VaclavObornik/mongodash.git
   *
   * Copyright (c) 2021 Václav Oborník
   * Released under the MIT license
   */
  "use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.cronTask = exports.scheduleCronTaskImmediately = exports.startCronTasks = exports.stopCronTasks = exports.init = exports.runCronTask = void 0;
// import * as _debug from 'debug';
const cron_parser_1 = require("cron-parser");
const parse_duration_1 = require("parse-duration");
const createContinuousLock_1 = require("./createContinuousLock");
const getCollection_1 = require("./getCollection");
const initPromise_1 = require("./initPromise");
class TaskDocument {
    constructor(_id, runSince) {
        this._id = _id;
        this.runSince = runSince;
        this.runImmediately = false;
        this.runLog = [];
        this.lockedTill = null;
    }
}
const noTaskWaitTime = 5 * 1000;
const state = {
    tasks: new Map(),
    nextTaskTimeoutId: null,
    working: false,
    _collection: null,
    get collection() {
        if (!this._collection) {
            this._collection = (0, getCollection_1.getCollection)('cronTasks');
        }
        return this._collection;
    },
    enforcedTasks: [],
    shouldRun: true,
    ensureIndexPromise: null,
    cronExpressionParserOptions: {},
};
let _onError;
function dateInMillis(milliseconds) {
    return new Date(Date.now() + milliseconds);
}
const invalidIntervalMessage = `Invalid interval.`;
function createIntervalFunctionFromScalar(interval) {
    if (typeof interval === 'number') {
        if (!Number.isFinite(interval)) {
            throw new Error('Interval number has to be finite.');
        }
        return () => dateInMillis(interval);
    }
    else if (typeof interval !== 'string') {
        throw new Error(invalidIntervalMessage);
    }
    if (/^CRON /i.test(interval)) {
        try {
            const expression = interval.slice(5);
            const parsedExpression = (0, cron_parser_1.parseExpression)(expression, state.cronExpressionParserOptions);
            return () => parsedExpression.next().toDate();
        }
        catch (err) {
            throw new Error(`${invalidIntervalMessage} ${err.message}.`);
        }
    }
    const duration = (0, parse_duration_1.default)(interval);
    // debug('parsed duration: ', duration);
    if (typeof duration !== 'number') {
        throw new Error(invalidIntervalMessage);
    }
    return () => dateInMillis(duration);
}
async function getNextRunDate(interval) {
    const maybeDate = await interval();
    if (maybeDate instanceof Date) {
        return maybeDate;
    }
    return createIntervalFunctionFromScalar(maybeDate)();
}
async function runCronTask(taskId) {
    var _a;
    if ((_a = new Error().stack) === null || _a === void 0 ? void 0 : _a.includes('mongoDashRunTaskNotCyclic')) {
        throw new Error('It is not possible to call runCronTask inside another running task. Use the scheduleCronTaskImmediately() function instead.');
    }
    // debug(`runCronTask called for ${taskId}`);
    if (!state.tasks.has(taskId)) {
        throw new Error(`Cannot run unknown task '${taskId}'.`);
    }
    return new Promise((resolve, reject) => {
        state.enforcedTasks.push({ taskId, resolve, reject });
        ensureStarted();
    });
}
exports.runCronTask = runCronTask;
function ensureIndex() {
    if (!state.ensureIndexPromise) {
        state.ensureIndexPromise = Promise.all([
            state.collection.createIndex({ runSince: 1, _id: 1, lockedTill: 1 }, { name: 'runSinceIndex' }),
            state.collection.createIndex({ runImmediately: 1, _id: 1, lockedTill: 1 }, { name: 'runImmediatelyIndex', partialFilterExpression: { runImmediately: { $eq: true } } }),
        ]);
    }
    return state.ensureIndexPromise;
}
const lockTime = 5 * 60 * 1000;
function getLockDate() {
    return new Date(Date.now() + lockTime);
}
function getUnlockedFilter() {
    return { $or: [{ lockedTill: null }, { lockedTill: { $lt: new Date() } }] };
}
function getTasksToProcessFilter() {
    return {
        $and: [{ _id: { $in: Array.from(state.tasks.keys()) } }, getUnlockedFilter()],
    };
}
async function findATaskToRun(enforcedTask) {
    let filter;
    if (enforcedTask) {
        filter = { $and: [{ _id: enforcedTask.taskId }, getUnlockedFilter()] };
    }
    else {
        filter = {
            $and: [{ $or: [{ runSince: { $lte: new Date() } }, { runImmediately: true }] }, getTasksToProcessFilter()],
        };
    }
    // debug('finding a task', JSON.stringify(filter, null, 2));
    const { value: document } = await state.collection.findOneAndUpdate(filter, {
        $set: {
            lockedTill: getLockDate(),
            runImmediately: false,
        },
        $push: {
            runLog: {
                $each: [{ startedAt: new Date(), finishedAt: null, error: null }],
                $sort: { startedAt: -1 },
                $slice: 5,
            },
        },
    }, {
        sort: {
            runImmediately: -1,
            runSince: 1,
            'runLog.0.finishedAt': 1, // prefer tasks waiting longer
        },
        projection: { _id: true, runImmediately: true },
    });
    if (!document) {
        if (enforcedTask) {
            enforcedTask.reject(new Error('The task document not found or is locked right now.'));
        }
        return null;
    }
    if (!enforcedTask && !state.shouldRun) {
        // the stopCronTasks has been called during finding a task, rollback the lock
        // we update runImmediately back only if it was truthy before
        const runImmediatelyUpdate = document.runImmediately ? { runImmediately: true } : null;
        await state.collection.updateOne({ _id: document._id }, {
            $set: Object.assign({ lockedTill: null }, runImmediatelyUpdate),
            $pop: { runLog: -1 }, // remove last runLog entry (0 index)
        });
        return null;
    }
    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    return state.tasks.get(document._id);
}
async function processTask(task, enforcedTask) {
    // debug(`processing task ${task.taskId}`);
    let taskError = null;
    // todo solve the "any"
    const stopContinuousLock = (0, createContinuousLock_1.createContinuousLock)(state.collection, task.taskId, 'lockedTill', lockTime, _onError);
    try {
        await (function mongoDashRunTaskNotCyclic() {
            return task.task();
        })();
    }
    catch (err) {
        taskError = err;
    }
    try {
        await stopContinuousLock(); // to avoid possibility of lock after the following document update
        const nextRunDate = await getNextRunDate(task.interval);
        // debug(`scheduling task ${task.taskId} to run in ${nextRunDate.getTime() - Date.now()}ms`);
        await state.collection.updateOne({ _id: task.taskId }, {
            $set: {
                runSince: nextRunDate,
                lockedTill: null,
                'runLog.0.error': taskError ? `${taskError}` : null,
                'runLog.0.finishedAt': new Date(),
            },
        });
    }
    finally {
        if (enforcedTask) {
            if (taskError) {
                enforcedTask.reject(taskError);
            }
            else {
                enforcedTask.resolve();
            }
        }
        else if (taskError) {
            _onError(taskError);
        }
    }
}
/** can never throw */
async function getWaitTimeByNextTask() {
    try {
        const nextTask = await state.collection.findOne(getTasksToProcessFilter(), {
            projection: { runSince: 1 },
            sort: { runSince: 1 },
        });
        if (!nextTask) {
            return noTaskWaitTime;
        }
        const timeToNext = nextTask.runSince.getTime() - Date.now();
        return Math.min(Math.max(timeToNext, 0), noTaskWaitTime);
    }
    catch (error) {
        _onError(error);
        return noTaskWaitTime;
    }
}
function runATask() {
    // debug('runATask called');
    state.working = true;
    (async () => {
        await initPromise_1.initPromise;
        const enforcedTask = state.enforcedTasks.shift() || null; // there can be no enforced task
        let task = null;
        const countOfTasks = state.tasks.size;
        try {
            task = await findATaskToRun(enforcedTask);
            if (!task) {
                // debug('no pending task found');
                return; // the finally statement will be called anyway
            }
            await processTask(task, enforcedTask);
        }
        catch (error) {
            // debug(`Catch error ${error}`);
            if (enforcedTask) {
                enforcedTask.reject(error);
            }
            else {
                _onError(error);
            }
        }
        finally {
            const shouldTriggerNext = () => state.shouldRun || !!state.enforcedTasks.length;
            if (shouldTriggerNext()) {
                // if there was no task, wait for standard time,
                // but try find another one when one has finished
                // or a new task has been registered
                const aTaskHasBeenRegistered = () => state.tasks.size !== countOfTasks;
                let waitTime = 0;
                if (!task && !aTaskHasBeenRegistered() && !state.enforcedTasks.length) {
                    waitTime = await getWaitTimeByNextTask();
                    // a task can be registered in the meanwhile
                    if (aTaskHasBeenRegistered() || state.enforcedTasks.length) {
                        waitTime = 0;
                    }
                }
                // should we still trigger next?
                if (shouldTriggerNext()) {
                    // debug(`SCHEDULING NEXT CHECK AFTER ${waitTime}ms`);
                    state.nextTaskTimeoutId = setTimeout(() => {
                        // debug("it's time!");
                        state.nextTaskTimeoutId = null;
                        runATask();
                    }, waitTime);
                }
            }
            state.working = false;
        }
    })();
}
function init({ runCronTasks, cronExpressionParserOptions, onError }) {
    if (cronExpressionParserOptions.endDate) {
        throw new Error("The 'endDate' parameter of the cron-parser package is not supported yet.");
    }
    state.shouldRun = runCronTasks;
    _onError = onError;
    state.cronExpressionParserOptions = cronExpressionParserOptions;
}
exports.init = init;
function ensureStarted() {
    // terminate waiting if there is any
    if (state.nextTaskTimeoutId) {
        clearTimeout(state.nextTaskTimeoutId);
        state.nextTaskTimeoutId = null;
    }
    if (!state.working) {
        // debug('STARTING LOOP');
        runATask();
    }
    // else the loop is already set
}
function stopCronTasks() {
    // debug('STOPPING CRON TASKS');
    state.shouldRun = false;
    if (state.nextTaskTimeoutId) {
        clearTimeout(state.nextTaskTimeoutId);
    }
}
exports.stopCronTasks = stopCronTasks;
function startCronTasks() {
    state.shouldRun = true;
    if (state.tasks.size) {
        ensureStarted();
    }
}
exports.startCronTasks = startCronTasks;
async function scheduleCronTaskImmediately(taskId) {
    const { matchedCount } = await state.collection.updateOne({ _id: taskId }, { $set: { runImmediately: true } });
    if (!matchedCount) {
        throw new Error(`No task with id "${taskId}" is registered.`);
    }
    if (state.shouldRun && state.tasks.has(taskId)) {
        ensureStarted();
    }
}
exports.scheduleCronTaskImmediately = scheduleCronTaskImmediately;
async function cronTask(taskId, interval, task) {
    await initPromise_1.initPromise;
    if (state.tasks.has(taskId)) {
        throw new Error(`The taskId '${taskId}' is already used.`);
    }
    const intervalFunction = typeof interval === 'function' ? interval : createIntervalFunctionFromScalar(interval);
    const nextRun = await getNextRunDate(intervalFunction);
    const document = new TaskDocument(taskId, nextRun);
    await state.collection.updateOne({ _id: document._id }, { $setOnInsert: document }, { upsert: true });
    state.tasks.set(taskId, {
        taskId,
        task,
        interval: intervalFunction,
    });
    // debug(`task ${taskId} has been registered`);
    await ensureIndex();
    if (state.shouldRun) {
        ensureStarted();
    }
}
exports.cronTask = cronTask;
//# sourceMappingURL=cronTasks.js.map