/**!
   * mongodash v1.1.1
   * https://github.com/VaclavObornik/mongodash.git
   *
   * Copyright (c) 2021 Václav Oborník
   * Released under the MIT license
   */
  'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
exports.withLock = exports.isLockAlreadyAcquiredError = exports.LockAlreadyAcquiredError = exports.init = void 0;
// import * as _debug from 'debug';
const mongodb_1 = require("mongodb");
const createContinuousLock_1 = require("./createContinuousLock");
const getCollection_1 = require("./getCollection");
const lodash_1 = require("lodash");
// const debug = _debug('mongodash:withLock');
let onError;
function init(options) {
    onError = options.onError;
}
exports.init = init;
const lockAcquiredMessage = 'The lock is already acquired.';
const expirationKey = 'expiresAt';
let collectionPromise;
async function getLockerCollection() {
    if (!collectionPromise) {
        collectionPromise = (async () => {
            const collection = (0, getCollection_1.getCollection)('locks');
            await collection.createIndex({ [expirationKey]: 1 }, { name: 'expiresAtIndex', expireAfterSeconds: 0 });
            return collection;
        })();
    }
    return collectionPromise;
}
class LockAlreadyAcquiredError extends Error {
    constructor(key) {
        super(lockAcquiredMessage);
        this.key = key;
    }
}
exports.LockAlreadyAcquiredError = LockAlreadyAcquiredError;
function isLockAlreadyAcquiredError(err, key) {
    const rightError = err instanceof LockAlreadyAcquiredError;
    if (!rightError) {
        return false;
    }
    return key ? err.key === key : true;
}
exports.isLockAlreadyAcquiredError = isLockAlreadyAcquiredError;
async function withLock(key, callback, { maxWaitForLock = 3 * 1000, startingDelay = 50, expireIn = 15 * 1000 } = {}) {
    const lockId = new mongodb_1.ObjectId();
    const stringKey = `${key}`;
    const maxDate = new Date(Date.now() + maxWaitForLock);
    // maxDate is lowered for a queryDurationEstimate because we will try to acquire the lock after last wait
    const collection = await getLockerCollection();
    const releaseLock = () => collection.deleteOne({ _id: stringKey, lockId });
    let lastAcquireDuration = 0;
    const acquireLock = async () => {
        // debug('acquireLock called');
        const now = new Date();
        try {
            const matcher = { _id: stringKey, [expirationKey]: { $lte: now } };
            const doc = { lockId, [expirationKey]: new Date(now.getTime() + expireIn) };
            await collection.replaceOne(matcher, doc, { upsert: true });
        }
        catch (err) {
            if ([11000, 11001].includes(err.code)) {
                throw new LockAlreadyAcquiredError(key);
            }
            await releaseLock(); // the lock could be possible acquired
            throw err;
        }
        finally {
            lastAcquireDuration = Date.now() - now.getTime();
            // debug(`lastAcquireDuration: ${lastAcquireDuration}`);
        }
    };
    const maxDelay = Math.max(startingDelay, maxWaitForLock / 3);
    for (let n = 0; n < Number.MAX_SAFE_INTEGER; n++) {
        try {
            await acquireLock();
            break;
        }
        catch (err) {
            const randomMultiplier = (0, lodash_1.random)(1, 1.2, true);
            let waitTime = Math.min(2 ** n * randomMultiplier * startingDelay, maxDelay);
            let nextTime = new Date(Date.now() + waitTime);
            const ultimateAttemptStart = new Date(maxDate.getTime() - lastAcquireDuration);
            const restTime = ultimateAttemptStart.getTime() - Date.now();
            if (nextTime >= ultimateAttemptStart && restTime >= startingDelay) {
                // debug(`optimizing last wait time`);
                waitTime = restTime;
                nextTime = ultimateAttemptStart;
            }
            // debug(`wait time ${waitTime} for ${n}`);
            if (isLockAlreadyAcquiredError(err, key) && nextTime <= ultimateAttemptStart) {
                await new Promise((resolve) => setTimeout(resolve, waitTime));
            }
            else {
                throw err;
            }
        }
    }
    // todo solve the "any"
    const stopContinuousLock = (0, createContinuousLock_1.createContinuousLock)(collection, stringKey, expirationKey, expireIn, onError);
    const cleanUp = () => Promise.all([stopContinuousLock(), releaseLock()]);
    let value;
    let err;
    try {
        value = await callback();
    }
    catch (error) {
        err = error;
    }
    await cleanUp();
    if (err) {
        throw err;
    }
    return value;
}
exports.withLock = withLock;
//# sourceMappingURL=withLock.js.map